"""
tests for pyesgf.logon module.

"""

import tempfile
import os
import os.path as op
import shutil

from pyesgf.logon import LogonManager, ESGF_CREDENTIALS

TEST_USER = 'pyesgf'
TEST_PASSWORD = 'RetSet1'
TEST_MYPROXY = 'pcmdi9.llnl.gov'
TEST_OPENID = 'https://pcmdi9.llnl.gov/esgf-idp/openid/pyesgf'

esgf_dir = None
test_data_dir = op.join(op.dirname(__file__), 'data')

def setup_module():
    global esgf_dir
    esgf_dir = tempfile.mkdtemp(prefix='pyesgf_tmp')

def teardown_module():
    if op.exists(esgf_dir):
        shutil.rmtree(esgf_dir)
        
def _load_creds(credentials_file=None, certificates_tarball=None):
    if credentials_file:
        shutil.copy(op.join(test_data_dir, credentials_file),
        op.join(esgf_dir, 'credentials.pem'))
    if certificates_tarball:
        os.system('cd %s ; tar zxf %s' %
                  (esgf_dir, op.join(test_data_dir,
                                     certificates_tarball)))

def _clear_creds():
    cred_path = op.join(esgf_dir, ESGF_CREDENTIALS)
    if op.exists(cred_path):
        os.remove(cred_path)

def _clear_certs():
    cert_path = op.join(esgf_dir, 'certificates')
    if op.exists(cert_path):
        shutil.rmtree(cert_path)

def test_no_logon():
    _clear_creds()
    lm = LogonManager(esgf_dir)
    assert lm.is_logged_on() == False

def test_expired():
    _load_creds('expired.pem')
    lm = LogonManager(esgf_dir)
    assert lm.state == lm.STATE_EXPIRED_CREDENTIALS

def test_logon():
    _clear_creds()
    _load_creds(certificates_tarball='pcmdi9-certs.tar.gz')
    lm = LogonManager(esgf_dir)
    lm.logon(TEST_USER, TEST_PASSWORD, TEST_MYPROXY)

    assert lm.is_logged_on()

def test_bootstrap():
    _clear_creds()
    lm = LogonManager(esgf_dir)
    lm.logon(TEST_USER, TEST_PASSWORD, TEST_MYPROXY, bootstrap=True)

    assert lm.is_logged_on()

def test_logoff():
    lm = LogonManager(esgf_dir)

    # Only re-logon if credentials are not valid
    if not lm.is_logged_on():
        lm.logon(TEST_USER, TEST_PASSWORD, TEST_MYPROXY, bootstrap=True)

    assert lm.is_logged_on()
    lm.logoff()

    assert not op.exists(op.join(esgf_dir, 'credentials.pem'))
    assert not lm.is_logged_on()
    assert lm.state == lm.STATE_NO_CREDENTIALS

def test_logon_openid():
    _clear_creds()
    _load_creds(certificates_tarball='pcmdi9-certs.tar.gz')
    lm = LogonManager(esgf_dir)
    lm.logon_with_openid(TEST_OPENID, TEST_PASSWORD, interactive=False)

    assert lm.is_logged_on()

    



"""
An interface to the `ESGF Search API`_


.. _`ESGF Search API`: http://esgf.org/wiki/ESGF_Search_API

"""

from .connection import SearchConnection
from .context import SearchContext
from .constraints import GeospatialConstraint, any_of, not_equals
from .results import ResultSet
from .consts import TYPE_DATASET, TYPE_FILE


#!TODO: ResultFormatter class.  process response json to specialise the result json.  Default is None
#!TODO: pipe results to new process.  Command-line interface.

    
    

    



"""
Convenience interface for defining service URL endpoints for ESGF nodes

"""

class ESGFNode(object):
    def __init__(self, base_url):
        # Strip '/' from url as necessary
        self.base_url = base_url.rstrip('/')

    @property
    def search_url(self):
        """
        Return the URL of the esg-search service.

        This URL is the prefix required for search and wget endpoints.
        """
        return '/'.join(self.base_url, 'esg-search')

    @property
    def ats_url(self):
        """
        Return the URL for the ESGF SAML AttributeService
        """
        return '/'.join([self.base_url,
                        'esgf-idp/saml/soap/secure/attributeService.htm'])

    @property
    def azs_url(self):
        """
        Return the URL for the ESGF SAML AuthorizationService.
        """
        return '/'.join([self.base_url,
                        'esgf-orp/saml/soap/secure/authorizationService.htm'])

    
        



"""
test_temporal_search.py
=======================

Uses CMIP5 to find HadGEM2-ES dataset ids that include data between 1960 and 1962.

Tests whether the "from_timestamp" and "to_timestamp" options are working in esgf-pyclient. 

"""

from pyesgf.search import SearchConnection


def test_temporal_search_CMIP5():
    conn = SearchConnection('http://esgf-index1.ceda.ac.uk/esg-search',
                        distrib=False)

    ctx1 = conn.new_context(project = "CMIP5", model = "HadGEM2-ES",
          time_frequency = "mon", realm = "atmos", ensemble = "r1i1p1", latest = True)

    ctx2 = conn.new_context(project = "CMIP5", model = "HadGEM2-ES",
          time_frequency = "mon", realm = "atmos", ensemble = "r1i1p1", latest = True,
          from_timestamp = "2100-12-30T23:23:59Z", to_timestamp = "2200-01-01T00:00:00Z")

    assert ctx2.hit_count < ctx1.hit_count 

def test_temporal_search_CORDEX():
    conn = SearchConnection("http://esgf-data.dkrz.de/esg-search", distrib=True)

    ctx1 = conn.new_context(project='CORDEX',
           from_timestamp="1990-01-01T12:00:00Z",
           to_timestamp="2100-12-31T12:00:00Z")

    ctx2 = conn.new_context(project='CORDEX',
           from_timestamp="2011-01-01T12:00:00Z",
           to_timestamp="2100-12-31T12:00:00Z")

    assert ctx2.hit_count < ctx1.hit_count



"""
Test downloading wget scripts

"""

from pyesgf.search import SearchConnection, not_equals

from .config import TEST_SERVICE

def test_download_script():
    conn = SearchConnection(TEST_SERVICE, distrib=False)
    ctx = conn.new_context(project='CMIP5', ensemble='r1i1p1', model='IPSL-CM5A-LR', realm='seaIce',
                           experiment='historicalGHG')
    script = ctx.get_download_script()

    assert '# ESG Federation download script' in script
    assert '# Search URL: %s' % TEST_SERVICE in script



"""
Test the Results classes

"""

import re
from urlparse import urlparse

from pyesgf.search.connection import SearchConnection

from .config import TEST_SERVICE

def test_result1():
    conn = SearchConnection(TEST_SERVICE, distrib=False)

    ctx = conn.new_context(project='CMIP5')
    results = ctx.search()

    r1 = results[0]
    assert re.match(r'cmip5\.output1\.MOHC\..+\|esgf-data2.ceda.ac.uk', r1.dataset_id)
    
def test_file_context():
    conn = SearchConnection(TEST_SERVICE, distrib=False)

    ctx = conn.new_context(project='CMIP5')
    results = ctx.search()

    r1 = results[0]
    f_ctx = r1.file_context()

    assert f_ctx.facet_constraints['dataset_id'] == r1.dataset_id

def test_file_list():
    conn = SearchConnection(TEST_SERVICE, distrib=False)

    ctx = conn.new_context(project='CMIP5')
    results = ctx.search()

    r1 = results[0]
    f_ctx = r1.file_context()

    file_results = f_ctx.search()
    f1 = file_results[0]

    ds_id, shard = r1.dataset_id.split('|')
    download_url = f1.download_url

    # Assumes dataset is published with DRS path.
    ds_subpath = ds_id.replace('.', '/')
    assert ds_subpath.lower() in download_url.lower()

def test_file_list2():
    conn = SearchConnection(TEST_SERVICE, distrib=False)

    ctx = conn.new_context(project='CMIP5')
    results = ctx.search()

    r1 = results[0]
    f_ctx = r1.file_context()

    file_results = f_ctx.search()
    for file_result in file_results:
        assert re.search(r'ds/.*\.nc', file_result.download_url)

def test_aggregations():
    conn = SearchConnection(TEST_SERVICE, distrib=False)

    ctx = conn.new_context(project='CMIP5')
    results = ctx.search()

    r1 = results[0]
    agg_ctx = r1.aggregation_context()

    agg_results = agg_ctx.search()
    agg1 = agg_results[0]

    ds_id, shard = r1.dataset_id.split('|')
    las_url = agg1.urls['LAS'][0][0]

    #!FIXME: A pretty dumb test for a correct aggregation
    assert '.aggregation' in las_url


def test_index_node():
    conn = SearchConnection(TEST_SERVICE, distrib=False)

    ctx = conn.new_context(project='CMIP5')
    results = ctx.search()

    r1 = results[0]
    service = urlparse(TEST_SERVICE)

    assert r1.index_node == service.hostname


def test_other_index_node():
    conn = SearchConnection(TEST_SERVICE, distrib=True)

    ctx = conn.new_context(project='CMIP5', institute='INM')
    results = ctx.search()

    r1 = results[0]
    service = urlparse(TEST_SERVICE)
    print 'index_node = %s' % r1.index_node
    
    assert r1.index_node is not None
    assert r1.index_node != service.hostname

def test_shards_constrain():
    # Test that a file-context constrains the shard list
    
    conn = SearchConnection(TEST_SERVICE, distrib=True)

    ctx = conn.new_context(project='CMIP5')
    results = ctx.search()

    r1 = results[0]
    f_ctx = r1.file_context()

    #!TODO: white-box test.  Refactor.
    query_dict = f_ctx._build_query()
    full_query = f_ctx.connection._build_query(query_dict, shards=f_ctx.shards)

    #!TODO: Force fail to see whether shards is passed through.
    q_shard = full_query['shards']
    # Check it isn't a ',' separated list
    assert ',' not in q_shard
    q_shard_host = q_shard.split(':')[0]
    assert q_shard_host == r1.json['index_node']

    # Now make the query to make sure it returns data from the right index_node
    f_results = f_ctx.search()
    f_r1 = results[0]
    assert f_r1.json['index_node'] == r1.json['index_node']


def test_shards_constrain2():
    # Regression test for issue #8 reported by ian.edmond@metoffice.gov.uk
    conn = SearchConnection('http://pcmdi9.llnl.gov/esg-search',distrib=True)

    ctx = conn.new_context(experiment='piControl', time_frequency='day', variable='pr', ensemble='r1i1p1')
    ctx = ctx.constrain(query='cmip5.output1.BCC.bcc-csm1-1-m.piControl.day.atmos.day.r1i1p1')
    s = ctx.search()

    ds = s[0]

    publicationDataset, server = ds.dataset_id.split('|')
    print publicationDataset, server, ds.json['replica']
    
    searchContext = ds.file_context()
    searchContext=searchContext.constrain(variable='pr')
    for j in searchContext.search():
        print j.download_url, j.checksum, j.checksum_type, j.size

def test_shards_constrain3():
    # Regression test for issue #8 reported by ian.edmond@metoffice.gov.uk
    conn = SearchConnection('http://pcmdi9.llnl.gov/esg-search',distrib=True)

    ctx = conn.new_context(query='cmip5.output1.CMCC.CMCC-CESM.historical.mon.atmos.Amon.r1i1p1.v20130416')
    s = ctx.search()

    ds = s[0]

    publicationDataset, server = ds.dataset_id.split('|')
    print publicationDataset, server, ds.json['replica']
    
    searchContext = ds.file_context()
    searchContext=searchContext.constrain(variable='pr')
    for j in searchContext.search():
        print j.download_url, j.checksum, j.checksum_type, j.size


def test_shards_constrain4():
    # Regression test for issue #8 reported by ian.edmond@metoffice.gov.uk
    conn = SearchConnection('http://pcmdi9.llnl.gov/esg-search',distrib=True)

    ctx = conn.new_context(query='cmip5.output1.BCC.bcc-csm1-1-m.historical.mon.atmos.Amon.r1i1p1.v20120709')
    s = ctx.search()

    ds = s[0]

    publicationDataset, server = ds.dataset_id.split('|')
    print publicationDataset, server, ds.json['replica']
    
    searchContext = ds.file_context()
    searchContext=searchContext.constrain(variable='tas')
    for j in searchContext.search():
        print j.download_url, j.checksum, j.checksum_type, j.size



from .consts import OPERATOR_NEQ

class GeospatialConstraint(object):
    """
    Class to encapsulate all geospatial constraints in the ESGF Search API
    """

    def __init__(self, lat=None, lon=None, bbox=None, location=None,
                 radius=None, polygon=None):
        self.lat, self.lon = lat, lon
        self.bbox = bbox
        self.location = location
        self.radius, self.polygon = radius, polygon

#-----------------------------------------------------------------------------
# Convenience functions

#!TODO: document constraint operators
        
def any_of(values):
    """
    Constrains to any of the specified values.
    This is a synonym for list(values).
    
    """
    return list(values)

def not_equals(value):
    return (OPERATOR_NEQ, value)



"""
Test SearchConnection object

"""

#!TODO: replace calls to the a live search service with a mock.
#!TODO: Test for HTTP proxies

import nose.tools as nt

from pyesgf.search.connection import SearchConnection
import pyesgf.search.exceptions as exc

from .config import TEST_SERVICE

def test_blank_query():
    conn = SearchConnection(TEST_SERVICE)
    json = conn.send_search({})

    assert json.keys() == [u'facet_counts', u'responseHeader', u'response']
    
def test_get_shard_list_fail():
    conn = SearchConnection(TEST_SERVICE, distrib=False)
    nt.assert_raises(exc.EsgfSearchException, conn.get_shard_list)

def test_get_shard_list():
    conn = SearchConnection(TEST_SERVICE, distrib=True)
    shards = conn.get_shard_list()
    #!NOTE: the exact shard list will change depending on the shard replication configuration
    #    on the test server
    assert 'esgf-index2.ceda.ac.uk' in shards
    # IPSL now replicates all non-local shards.  Just check it has a few shards
    assert len(shards['esgf-index2.ceda.ac.uk']) > 3
    
    
def test_url_fixing():
    conn1 = SearchConnection(TEST_SERVICE)
    conn2 = SearchConnection(TEST_SERVICE+'/')
    conn3 = SearchConnection(TEST_SERVICE+'///')
    conn4 = SearchConnection(TEST_SERVICE+'/search')
    conn5 = SearchConnection(TEST_SERVICE+'/search///')

    assert conn1.url == conn2.url == conn3.url == conn4.url == conn5.url



"""
Interfaces to ESGF Security APIs as defined in the `ESGF Security Interface`_

.. _`ESGF Security Interface`: https://github.com/ESGF/esgf.github.io/wiki/Security%7CInterfaceControlDocument

"""

NS = {
    'soap11': "http://schemas.xmlsoap.org/soap/envelope/",
    'saml2p': "urn:oasis:names:tc:SAML:2.0:protocol",
    'saml': "urn:oasis:names:tc:SAML:2.0:assertion",
}



#-----------------------------------------------------------------------------
# Constants

TYPE_DATASET = 'Dataset'
TYPE_FILE = 'File'
TYPE_AGGREGATION = 'Aggregation'
QUERY_KEYWORD_TYPES = ('system', 'facet', 'freetext', 'temporal', 'geospatial')
RESPONSE_FORMAT='application/solr+json'
DEFAULT_BATCH_SIZE = 50

OPERATOR_NEQ = 'not_equal'

SHARD_REXP = r'(?P<host>.*?):(?P<port>\d*)/solr(?P<suffix>.*)'



"""
Test the SearchContext class

"""

from pyesgf.search import SearchConnection, not_equals

from .config import TEST_SERVICE

def test_context_freetext():
    conn = SearchConnection(TEST_SERVICE)
    context = conn.new_context(query="temperature")

    assert context.freetext_constraint == "temperature"

def test_context_facets1():
    conn = SearchConnection(TEST_SERVICE)
    context = conn.new_context(project='cmip5')

    assert context.facet_constraints['project'] == 'cmip5'
    
def test_context_facets1():
    conn = SearchConnection(TEST_SERVICE)
    context = conn.new_context(project='CMIP5')

    context2 = context.constrain(model="IPSL-CM5A-LR")
    assert context2.facet_constraints['project'] == 'CMIP5'
    assert context2.facet_constraints['model'] == 'IPSL-CM5A-LR'

def test_context_facets_multivalue():
    conn = SearchConnection(TEST_SERVICE)
    context = conn.new_context(project='CMIP5')

    context2 = context.constrain(model=['IPSL-CM5A-LR', 'IPSL-CM5A-MR'])
    assert context2.hit_count > 0
    
    assert context2.facet_constraints['project'] == 'CMIP5'
    assert sorted(context2.facet_constraints.getall('model')) == ['IPSL-CM5A-LR', 'IPSL-CM5A-MR']

def test_context_facet_multivalue2():
    conn = SearchConnection(TEST_SERVICE)
    context = conn.new_context(project='CMIP5', model='IPSL-CM5A-MR')
    assert context.facet_constraints.getall('model') == ['IPSL-CM5A-MR']

    
    context2 = context.constrain(model=['IPSL-CM5A-MR', 'IPSL-CM5A-LR'])
    assert sorted(context2.facet_constraints.getall('model')) == ['IPSL-CM5A-LR', 'IPSL-CM5A-MR']


def test_context_facet_multivalue3():
    conn = SearchConnection(TEST_SERVICE)
    ctx = conn.new_context(project='CMIP5', query='humidity', experiment='rcp45')
    hits1 = ctx.hit_count
    assert hits1 > 0
    ctx2 = conn.new_context(project='CMIP5', query='humidity',
                           experiment=['rcp45','rcp85'])
    hits2 = ctx2.hit_count

    assert hits2 > hits1


def test_context_facet_options():
    conn = SearchConnection(TEST_SERVICE)
    context = conn.new_context(project='CMIP5', model='IPSL-CM5A-LR',
                               ensemble='r1i1p1', experiment='rcp60',
                               realm='seaIce')

    assert context.get_facet_options().keys() == ['data_node', 'cf_standard_name', 'variable_long_name', 
                               'cmor_table', 'time_frequency', 'variable']


def test_context_facets3():
    conn = SearchConnection(TEST_SERVICE)
    
    context = conn.new_context(project='CMIP5')
    context2 = context.constrain(model="IPSL-CM5A-LR")

    results = context2.search()
    result = results[0]

    assert result.json['project'] == ['CMIP5']
    assert result.json['model'] == ['IPSL-CM5A-LR']

def test_facet_count():
    conn = SearchConnection(TEST_SERVICE)
    
    context = conn.new_context(project='CMIP5')
    context2 = context.constrain(model="IPSL-CM5A-LR")

    counts = context2.facet_counts
    assert counts['model'].keys() == ['IPSL-CM5A-LR']
    assert counts['project'].keys() == ['CMIP5']

def test_distrib():
    conn = SearchConnection(TEST_SERVICE, distrib=False)

    context = conn.new_context(project='CMIP5')
    count1 = context.hit_count

    conn2 = SearchConnection(TEST_SERVICE, distrib=True)
    context = conn2.new_context(project='CMIP5')
    count2 = context.hit_count

    assert count1 < count2

def test_constrain():
    conn = SearchConnection(TEST_SERVICE)
    
    context = conn.new_context(project='CMIP5')
    count1 = context.hit_count
    context = context.constrain(model="IPSL-CM5A-LR")
    count2 = context.hit_count

    assert count1 > count2


def test_constrain_freetext():
    conn = SearchConnection(TEST_SERVICE)

    context = conn.new_context(project='CMIP5', query='humidity')
    assert context.freetext_constraint == 'humidity'

    context = context.constrain(experiment='historical')
    assert context.freetext_constraint == 'humidity'

def test_constrain_regression1():
    conn = SearchConnection(TEST_SERVICE)

    context = conn.new_context(project='CMIP5', model='IPSL-CM5A-LR')
    assert 'experiment' not in context.facet_constraints

    context2 = context.constrain(experiment='historical')
    assert 'experiment' not in context.facet_constraints

def test_negative_facet():
    conn = SearchConnection(TEST_SERVICE)

    context = conn.new_context(project='CMIP5', model='IPSL-CM5A-LR')
    hits1 = context.hit_count

    print context.facet_counts['experiment']

    context2 = context.constrain(experiment='historical')
    hits2 = context2.hit_count

    context3 = context.constrain(experiment=not_equals('historical'))
    hits3 = context3.hit_count


    assert hits1 == hits2 + hits3
    
def test_replica():
    # Test that we can exclude replicas
    # This tests assumes the test dataset is replicated
    conn = SearchConnection(TEST_SERVICE)

    context = conn.new_context(
        query='cmip5.output1.NIMR-KMA.HadGEM2-AO.rcp60.mon.atmos.Amon.r1i1p1')

    assert context.hit_count > 1

    context = conn.new_context(
        query='cmip5.output1.NIMR-KMA.HadGEM2-AO.rcp60.mon.atmos.Amon.r1i1p1',
        replica=False)

    assert context.hit_count == 1

def test_response_from_bad_parameter():
    # Test that a bad parameter name raises a useful exception
    # NOTE::: !!! This fails because urllib2 HTTP query is overrided with 
    #         !!! cache handler instead of usual response. 
    #         !!! Fix needs to make sure cached URL request has response exceptions matching urllib2 exception
    conn = SearchConnection(TEST_SERVICE)
    context = conn.new_context(project='CMIP5', rubbish='nonsense')
    context.hit_count

    try:
        context.hit_count
    except Exception, err:
        assert str(err).strip() == "Invalid query parameter(s): rubbish"
        




"""
Test Attribute Service API.

"""

from pyesgf.security.ats import AttributeService
from pyesgf.node import ESGFNode

CEDA_NODE = ESGFNode('https://esgf-index1.ceda.ac.uk')

def test_ceda_ats():
    service = AttributeService(CEDA_NODE.ats_url, 'esgf-pyclient')
    openid = 'https://ceda.ac.uk/openid/Stephen.Pascoe'
    resp = service.send_request(openid, ['urn:esg:first:name', 'urn:esg:last:name'])

    assert resp.get_subject() == openid
    
    attrs = resp.get_attributes()
    assert attrs['urn:esg:first:name'] == 'Stephen'
    assert attrs['urn:esg:last:name'] == 'Pascoe'
    
def test_unkown_principal():
    service = AttributeService(CEDA_NODE.ats_url, 'esgf-pyclient')
    openid = 'https://example.com/unknown'

    resp = service.send_request(openid, [])

    assert resp.get_status() == 'urn:oasis:names:tc:SAML:2.0:status:UnknownPrincipal'

def test_multi_attribute():
    service = AttributeService(CEDA_NODE.ats_url, 'esgf-pyclient')

    openid = 'https://ceda.ac.uk/openid/Stephen.Pascoe'
    resp = service.send_request(openid, ['CMIP5 Research'])

    attrs = resp.get_attributes()
    assert list(sorted(attrs['CMIP5 Research'])) == ['default', 'user']



# (c) 2005 Ian Bicking and contributors; written for Paste (http://pythonpaste.org)
# Licensed under the MIT license: http://www.opensource.org/licenses/mit-license.php
#
# Taken verbatim from https://bitbucket.org/ianb/webob/src/tip/webob/multidict.py
"""
Gives a multi-value dictionary object (MultiDict) plus several wrappers
"""
import cgi, copy, sys, warnings, urllib
from UserDict import DictMixin


__all__ = ['MultiDict', 'UnicodeMultiDict', 'NestedMultiDict', 'NoVars',
           'TrackableMultiDict']

class MultiDict(DictMixin):
    """
        An ordered dictionary that can have multiple values for each key.
        Adds the methods getall, getone, mixed and extend and add to the normal
        dictionary interface.
    """

    def __init__(self, *args, **kw):
        if len(args) > 1:
            raise TypeError("MultiDict can only be called with one positional argument")
        if args:
            if hasattr(args[0], 'iteritems'):
                items = list(args[0].iteritems())
            elif hasattr(args[0], 'items'):
                items = args[0].items()
            else:
                items = list(args[0])
            self._items = items
        else:
            self._items = []
        if kw:
            self._items.extend(kw.iteritems())

    @classmethod
    def view_list(cls, lst):
        """
        Create a dict that is a view on the given list
        """
        if not isinstance(lst, list):
            raise TypeError(
                "%s.view_list(obj) takes only actual list objects, not %r"
                % (cls.__name__, lst))
        obj = cls()
        obj._items = lst
        return obj

    @classmethod
    def from_fieldstorage(cls, fs):
        """
        Create a dict from a cgi.FieldStorage instance
        """
        obj = cls()
        # fs.list can be None when there's nothing to parse
        for field in fs.list or ():
            if field.filename:
                obj.add(field.name, field)
            else:
                obj.add(field.name, field.value)
        return obj

    def __getitem__(self, key):
        for k, v in reversed(self._items):
            if k == key:
                return v
        raise KeyError(key)

    def __setitem__(self, key, value):
        try:
            del self[key]
        except KeyError:
            pass
        self._items.append((key, value))

    def add(self, key, value):
        """
        Add the key and value, not overwriting any previous value.
        """
        self._items.append((key, value))

    def getall(self, key):
        """
        Return a list of all values matching the key (may be an empty list)
        """
        result = []
        for k, v in self._items:
            if key == k:
                result.append(v)
        return result

    def getone(self, key):
        """
        Get one value matching the key, raising a KeyError if multiple
        values were found.
        """
        v = self.getall(key)
        if not v:
            raise KeyError('Key not found: %r' % key)
        if len(v) > 1:
            raise KeyError('Multiple values match %r: %r' % (key, v))
        return v[0]

    def mixed(self):
        """
        Returns a dictionary where the values are either single
        values, or a list of values when a key/value appears more than
        once in this dictionary.  This is similar to the kind of
        dictionary often used to represent the variables in a web
        request.
        """
        result = {}
        multi = {}
        for key, value in self.iteritems():
            if key in result:
                # We do this to not clobber any lists that are
                # *actual* values in this dictionary:
                if key in multi:
                    result[key].append(value)
                else:
                    result[key] = [result[key], value]
                    multi[key] = None
            else:
                result[key] = value
        return result

    def dict_of_lists(self):
        """
        Returns a dictionary where each key is associated with a list of values.
        """
        r = {}
        for key, val in self.iteritems():
            r.setdefault(key, []).append(val)
        return r

    def __delitem__(self, key):
        items = self._items
        found = False
        for i in range(len(items)-1, -1, -1):
            if items[i][0] == key:
                del items[i]
                found = True
        if not found:
            raise KeyError(key)

    def __contains__(self, key):
        for k, v in self._items:
            if k == key:
                return True
        return False

    has_key = __contains__

    def clear(self):
        self._items = []

    def copy(self):
        return self.__class__(self)

    def setdefault(self, key, default=None):
        for k, v in self._items:
            if key == k:
                return v
        self._items.append((key, default))
        return default

    def pop(self, key, *args):
        if len(args) > 1:
            raise TypeError, "pop expected at most 2 arguments, got "\
                              + repr(1 + len(args))
        for i in range(len(self._items)):
            if self._items[i][0] == key:
                v = self._items[i][1]
                del self._items[i]
                return v
        if args:
            return args[0]
        else:
            raise KeyError(key)

    def popitem(self):
        return self._items.pop()

    def update(self, *args, **kw):
        if args:
            lst = args[0]
            if len(lst) != len(dict(lst)):
                # this does not catch the cases where we overwrite existing
                # keys, but those would produce too many warning
                msg = ("Behavior of MultiDict.update() has changed "
                    "and overwrites duplicate keys. Consider using .extend()"
                )
                warnings.warn(msg, UserWarning, stacklevel=2)
        DictMixin.update(self, *args, **kw)

    def extend(self, other=None, **kwargs):
        if other is None:
            pass
        elif hasattr(other, 'items'):
            self._items.extend(other.items())
        elif hasattr(other, 'keys'):
            for k in other.keys():
                self._items.append((k, other[k]))
        else:
            for k, v in other:
                self._items.append((k, v))
        if kwargs:
            self.update(kwargs)

    def __repr__(self):
        items = map('(%r, %r)'.__mod__, _hide_passwd(self.iteritems()))
        return '%s([%s])' % (self.__class__.__name__, ', '.join(items))

    def __len__(self):
        return len(self._items)

    ##
    ## All the iteration:
    ##

    def keys(self):
        return [k for k, v in self._items]

    def iterkeys(self):
        for k, v in self._items:
            yield k

    __iter__ = iterkeys

    def items(self):
        return self._items[:]

    def iteritems(self):
        return iter(self._items)

    def values(self):
        return [v for k, v in self._items]

    def itervalues(self):
        for k, v in self._items:
            yield v

class UnicodeMultiDict(DictMixin):
    """
    A MultiDict wrapper that decodes returned values to unicode on the
    fly. Decoding is not applied to assigned values.

    The key/value contents are assumed to be ``str``/``strs`` or
    ``str``/``FieldStorages`` (as is returned by the ``paste.request.parse_``
    functions).

    Can optionally also decode keys when the ``decode_keys`` argument is
    True.

    ``FieldStorage`` instances are cloned, and the clone's ``filename``
    variable is decoded. Its ``name`` variable is decoded when ``decode_keys``
    is enabled.

    """
    def __init__(self, multi, encoding=None, errors='strict',
                 decode_keys=False):
        self.multi = multi
        if encoding is None:
            encoding = sys.getdefaultencoding()
        self.encoding = encoding
        self.errors = errors
        self.decode_keys = decode_keys

    def _decode_key(self, key):
        if self.decode_keys:
            try:
                key = key.decode(self.encoding, self.errors)
            except AttributeError:
                pass
        return key

    def _encode_key(self, key):
        if self.decode_keys and isinstance(key, unicode):
            return key.encode(self.encoding, self.errors)
        return key

    def _decode_value(self, value):
        """
        Decode the specified value to unicode. Assumes value is a ``str`` or
        `FieldStorage`` object.

        ``FieldStorage`` objects are specially handled.
        """
        if isinstance(value, cgi.FieldStorage):
            # decode FieldStorage's field name and filename
            value = copy.copy(value)
            if self.decode_keys:
                if not isinstance(value.name, unicode):
                    value.name = value.name.decode(self.encoding, self.errors)
            if value.filename:
                if not isinstance(value.filename, unicode):
                    value.filename = value.filename.decode(self.encoding,
                                                           self.errors)
        elif not isinstance(value, unicode):
            try:
                value = value.decode(self.encoding, self.errors)
            except AttributeError:
                pass
        return value

    def _encode_value(self, value):
        if isinstance(value, unicode):
            value = value.encode(self.encoding, self.errors)
        return value

    def __getitem__(self, key):
        return self._decode_value(self.multi.__getitem__(self._encode_key(key)))

    def __setitem__(self, key, value):
        self.multi.__setitem__(self._encode_key(key), self._encode_value(value))

    def add(self, key, value):
        """
        Add the key and value, not overwriting any previous value.
        """
        self.multi.add(self._encode_key(key), self._encode_value(value))

    def getall(self, key):
        """
        Return a list of all values matching the key (may be an empty list)
        """
        return map(self._decode_value, self.multi.getall(self._encode_key(key)))

    def getone(self, key):
        """
        Get one value matching the key, raising a KeyError if multiple
        values were found.
        """
        return self._decode_value(self.multi.getone(self._encode_key(key)))

    def mixed(self):
        """
        Returns a dictionary where the values are either single
        values, or a list of values when a key/value appears more than
        once in this dictionary.  This is similar to the kind of
        dictionary often used to represent the variables in a web
        request.
        """
        unicode_mixed = {}
        for key, value in self.multi.mixed().iteritems():
            if isinstance(value, list):
                value = [self._decode_value(value) for value in value]
            else:
                value = self._decode_value(value)
            unicode_mixed[self._decode_key(key)] = value
        return unicode_mixed

    def dict_of_lists(self):
        """
        Returns a dictionary where each key is associated with a
        list of values.
        """
        unicode_dict = {}
        for key, value in self.multi.dict_of_lists().iteritems():
            value = [self._decode_value(value) for value in value]
            unicode_dict[self._decode_key(key)] = value
        return unicode_dict

    def __delitem__(self, key):
        self.multi.__delitem__(self._encode_key(key))

    def __contains__(self, key):
        return self.multi.__contains__(self._encode_key(key))

    has_key = __contains__

    def clear(self):
        self.multi.clear()

    def copy(self):
        return UnicodeMultiDict(self.multi.copy(), self.encoding, self.errors)

    def setdefault(self, key, default=None):
        return self._decode_value(
            self.multi.setdefault(self._encode_key(key),
                                  self._encode_value(default)))

    def pop(self, key, *args):
        return self._decode_value(self.multi.pop(self._encode_key(key), *args))

    def popitem(self):
        k, v = self.multi.popitem()
        return (self._decode_key(k), self._decode_value(v))

    def __repr__(self):
        items = map('(%r, %r)'.__mod__, _hide_passwd(self.iteritems()))
        return '%s([%s])' % (self.__class__.__name__, ', '.join(items))

    def __len__(self):
        return self.multi.__len__()

    ##
    ## All the iteration:
    ##

    def keys(self):
        return [self._decode_key(k) for k in self.multi.iterkeys()]

    def iterkeys(self):
        for k in self.multi.iterkeys():
            yield self._decode_key(k)

    __iter__ = iterkeys

    def items(self):
        return [(self._decode_key(k), self._decode_value(v))
                for k, v in self.multi.iteritems()]

    def iteritems(self):
        for k, v in self.multi.iteritems():
            yield (self._decode_key(k), self._decode_value(v))

    def values(self):
        return [self._decode_value(v) for v in self.multi.itervalues()]

    def itervalues(self):
        for v in self.multi.itervalues():
            yield self._decode_value(v)

_dummy = object()

class TrackableMultiDict(MultiDict):
    tracker = None
    name = None
    def __init__(self, *args, **kw):
        if '__tracker' in kw:
            self.tracker = kw.pop('__tracker')
        if '__name' in kw:
            self.name = kw.pop('__name')
        MultiDict.__init__(self, *args, **kw)
    def __setitem__(self, key, value):
        MultiDict.__setitem__(self, key, value)
        self.tracker(self, key, value)
    def add(self, key, value):
        MultiDict.add(self, key, value)
        self.tracker(self, key, value)
    def __delitem__(self, key):
        MultiDict.__delitem__(self, key)
        self.tracker(self, key)
    def clear(self):
        MultiDict.clear(self)
        self.tracker(self)
    def setdefault(self, key, default=None):
        result = MultiDict.setdefault(self, key, default)
        self.tracker(self, key, result)
        return result
    def pop(self, key, *args):
        result = MultiDict.pop(self, key, *args)
        self.tracker(self, key)
        return result
    def popitem(self):
        result = MultiDict.popitem(self)
        self.tracker(self)
        return result
    def update(self, *args, **kwargs):
        MultiDict.update(self, *args, **kwargs)
        self.tracker(self)
    def __repr__(self):
        items = map('(%r, %r)'.__mod__, _hide_passwd(self.iteritems()))
        return '%s([%s])' % (self.name or self.__class__.__name__, ', '.join(items))
    def copy(self):
        # Copies shouldn't be tracked
        return MultiDict(self)

class NestedMultiDict(MultiDict):
    """
    Wraps several MultiDict objects, treating it as one large MultiDict
    """

    def __init__(self, *dicts):
        self.dicts = dicts

    def __getitem__(self, key):
        for d in self.dicts:
            value = d.get(key, _dummy)
            if value is not _dummy:
                return value
        raise KeyError(key)

    def _readonly(self, *args, **kw):
        raise KeyError("NestedMultiDict objects are read-only")
    __setitem__ = _readonly
    add = _readonly
    __delitem__ = _readonly
    clear = _readonly
    setdefault = _readonly
    pop = _readonly
    popitem = _readonly
    update = _readonly

    def getall(self, key):
        result = []
        for d in self.dicts:
            result.extend(d.getall(key))
        return result

    # Inherited:
    # getone
    # mixed
    # dict_of_lists

    def copy(self):
        return MultiDict(self)

    def __contains__(self, key):
        for d in self.dicts:
            if key in d:
                return True
        return False

    has_key = __contains__

    def __len__(self):
        v = 0
        for d in self.dicts:
            v += len(d)
        return v

    def __nonzero__(self):
        for d in self.dicts:
            if d:
                return True
        return False

    def items(self):
        return list(self.iteritems())

    def iteritems(self):
        for d in self.dicts:
            for item in d.iteritems():
                yield item

    def values(self):
        return list(self.itervalues())

    def itervalues(self):
        for d in self.dicts:
            for value in d.itervalues():
                yield value

    def keys(self):
        return list(self.iterkeys())

    def __iter__(self):
        for d in self.dicts:
            for key in d:
                yield key

    iterkeys = __iter__

class NoVars(object):
    """
    Represents no variables; used when no variables
    are applicable.

    This is read-only
    """

    def __init__(self, reason=None):
        self.reason = reason or 'N/A'

    def __getitem__(self, key):
        raise KeyError("No key %r: %s" % (key, self.reason))

    def __setitem__(self, *args, **kw):
        raise KeyError("Cannot add variables: %s" % self.reason)

    add = __setitem__
    setdefault = __setitem__
    update = __setitem__

    def __delitem__(self, *args, **kw):
        raise KeyError("No keys to delete: %s" % self.reason)
    clear = __delitem__
    pop = __delitem__
    popitem = __delitem__

    def get(self, key, default=None):
        return default

    def getall(self, key):
        return []

    def getone(self, key):
        return self[key]

    def mixed(self):
        return {}
    dict_of_lists = mixed

    def __contains__(self, key):
        return False
    has_key = __contains__

    def copy(self):
        return self

    def __repr__(self):
        return '<%s: %s>' % (self.__class__.__name__,
                             self.reason)

    def __len__(self):
        return 0

    def __cmp__(self, other):
        return cmp({}, other)

    def keys(self):
        return []
    def iterkeys(self):
        return iter([])
    __iter__ = iterkeys
    items = keys
    iteritems = iterkeys
    values = keys
    itervalues = iterkeys



def _hide_passwd(items):
    for k, v in items:
        if ('password' in k
            or 'passwd' in k
            or 'pwd' in k
        ):
            yield k, '******'
        else:
            yield k, v



"""
test for secure opendap support

"""

import tempfile
import os
import os.path as op
import shutil
from unittest import TestCase
import re

from pyesgf.logon import LogonManager, DAP_CONFIG_MARKER
from pyesgf.search import SearchConnection
from test.config import TEST_SERVICE

class TestOpendap(TestCase):
    def setUp(self):
        self.esgf_dir = tempfile.mkdtemp(prefix='pyesgf_tmp')
        self.dap_config = op.join(self.esgf_dir, '.dodsrc')
        # NetCDF DAP support looks in CWD for configuration
        self.orig_dir = os.getcwd()
        os.chdir(self.esgf_dir)

    def tearDown(self):
        os.chdir(self.orig_dir)
        if op.exists(self.esgf_dir):
            shutil.rmtree(self.esgf_dir)

    def init_config(self, config):
        with open(self.dap_config, 'w') as fh:
            fh.write(config)

    def read_config(self):
        return open(self.dap_config).read()


    def check_preamble(self, preamble, config):
        return re.match((preamble + 
                         r'\s*^# BEGIN {0}'.format(DAP_CONFIG_MARKER)), 
                        config, re.M | re.S)

    def check_postamble(self, postamble, config):
        return re.search(r'^# END {0}$\s*{1}'.format(DAP_CONFIG_MARKER, postamble),
                         config, re.M | re.S)
    
    def test_config1(self):
        # Create the config file from scratch
        lm = LogonManager(self.esgf_dir, dap_config=self.dap_config)
        config = self.read_config()
        
        print config
        assert re.match(r'\s*^# BEGIN {0}$.*^# END {0}$'.format(DAP_CONFIG_MARKER),
                        config, re.M | re.S)

    def test_config2(self):
        # Create the config when one already exists.  Check it is retained.
        lines = ['# Welcome to my config file', 'SOME_OPT=foo', '']
        preamble = '\n'.join(lines)
        self.init_config(preamble)

        lm = LogonManager(self.esgf_dir, dap_config=self.dap_config)
        config = self.read_config()

        print config
        assert self.check_preamble(preamble, config)

    def test_config3(self):
        # Create the config when one already exists with the BEGIN section in it

        lines = ['# Welcome to my config file', 'SOME_OPT=foo', '']
        preamble = '\n'.join(lines)
                        
        lines = ['', '# Some more config here', 'OTHER_OPT=bar', '']
        postamble = '\n'.join(lines)

        config = '''\
{0}
# BEGIN <<< Managed by esgf-pyclient >>>
CURL.VERBOSE=0
CURL.COOKIEJAR=/tmp/foo/certificates/.dods_cookies
CURL.SSL.VALIDATE=1
CURL.SSL.CERTIFICATE=/tmp/foo/certificates/credentials.pem
CURL.SSL.KEY=/tmp/foo/certificates/credentials.pem
CURL.SSL.CAPATH=/tmp/foo/certificates/certificates
# END <<< Managed by esgf-pyclient >>>

{1}
'''.format(preamble, postamble)
                        
        self.init_config(config)

        lm = LogonManager(self.esgf_dir, dap_config=self.dap_config)
        config1 = self.read_config()
        
        print config1
        assert self.check_preamble(preamble, config1)
        assert self.check_postamble(postamble, config1)
    

    def test_open_url(self):
        import netCDF4

        lm = LogonManager(self.esgf_dir, dap_config=self.dap_config)
        print 'Using dap_config at %s' % self.dap_config

        conn = SearchConnection(TEST_SERVICE, distrib=False)

        #!TODO: replace with request for specific dataset
        ctx = conn.new_context(project='CMIP5')
        results = ctx.search()

        r1 = results[0]
        f_ctx = r1.file_context()

        file_results = f_ctx.search()
        
        opendap_url = file_results[0].opendap_url
        print 'OPeNDAP URL is %s' % opendap_url

        ds = netCDF4.Dataset(opendap_url)
        print ds.variables.keys()
    test_open_url.__test__ = False


#!TODO: more corner cases to test for in DAP_CONFIG



"""

Module :mod:`pyesgf.search.context`
===================================

Defines the :class:`SearchContext` class which represents each ESGF search query.

"""

import copy

from pyesgf.multidict import MultiDict

from .constraints import GeospatialConstraint
from .consts import (TYPE_DATASET, TYPE_FILE, TYPE_AGGREGATION,
                     QUERY_KEYWORD_TYPES)
from .results import ResultSet
from .exceptions import EsgfSearchException


class SearchContext(object):
    """
    Instances of this class represent the state of a current search.
    It exposes what facets are available to select and the facet counts
    if they are available.
    
    Subclasses of this class can restrict the search options.  For instance
    FileSearchContext, DatasetSerachContext or CMIP5SearchContext
    
    SearchContext instances are connected to SearchConnection instances.  You
    normally create SearchContext instances via one of:
    1. Calling SearchConnection.new_context()
    2. Calling SearchContext.constrain()
    
    :ivar constraints: A dictionary of facet constraints currently in effect.
        ``constraint[facet_name] = [value, value, ...]``
    :property facet_counts: A dictionary of available hits with each 
        facet value for the search as currently constrained.
        This property returns a dictionary of dictionaries where 
        ``facet_counts[facet][facet_value] == hit_count``
    :property hit_count: The total number of hits available with current constraints.

        
    """

    DEFAULT_SEARCH_TYPE = NotImplemented

    def __init__(self, connection, constraints, search_type=None,
                 latest=None, facets=None, fields=None,
                 from_timestamp=None, to_timestamp=None,
                 replica=None, shards=None):
        """

        :param connection: The SearchConnection
        :param constraints: A dictionary of initial constraints
        :param search_type: One of TYPE_* constants defining the document
            type to search for.  Overrides SearchContext.DEFAULT_SEARCH_TYPE
        :param facets: The list of facets for which counts will be retrieved
            and constraints be validated against.  Or None to represent all
            facets.
        :param fields: A list of field names to return in search responses
        :param replica: A boolean defining whether to return master records
            or replicas, or None to return both.
        :param latest: A boolean defining whether to return only latest verisons
            or only non-latest versions, or None to return both.
        :param shards: list of shards to restrict searches to.  Should be from the list
            self.connection.get_shard_list()
        :param from_timestamp: Date-time string to specify start of search range 
            (e.g. "2000-01-01T00:00:00Z"). 
        :param to_timestamp: Date-time string to specify end of search range
            (e.g. "2100-12-31T23:59:59Z").

        """

        self.connection = connection
        self.__facet_counts = None
        self.__hit_count = None

        if search_type is None:
            search_type = self.DEFAULT_SEARCH_TYPE

        #  Constraints
        self.freetext_constraint = None
        self.facet_constraints = MultiDict()
        self.temporal_constraint = [from_timestamp, to_timestamp]
        self.geosplatial_constraint = None

        self._update_constraints(constraints)

        # Search configuration parameters
        self.timestamp_range = (from_timestamp, to_timestamp)

        search_types = [TYPE_DATASET, TYPE_FILE, TYPE_AGGREGATION]
        if search_type not in search_types:
            raise EsgfSearchException('search_type must be one of %s'
                                      % ','.join(search_types))
        self.search_type = search_type

        self.latest = latest
        self.facets = facets
        self.fields = fields
        self.replica = replica
        self.shards = shards

    #-------------------------------------------------------------------------
    # Functional search interface
    # These do not change the constraints on self.

    def search(self, **constraints):
        """
        Perform the search with current constraints returning a set of results.

        :param constraints: Further constraints for this query.  Equivilent
            to calling self.constrain(**constraints).search()
        :return: A ResultSet for this query

        """
        if constraints:
            sc = self.constrain(**constraints)
        else:
            sc = self

        self.__update_counts()

        return ResultSet(sc)

    def constrain(self, **constraints):
        """
        Return a *new* instance with the additional constraints.
        
        """
        new_sc = copy.deepcopy(self)
        new_sc._update_constraints(constraints)
        return new_sc

    def get_download_script(self, **constraints):
        """
        Download a script for downloading all files in the set of results.

        :param constraints: Further constraints for this query.  Equivilent
            to calling self.constrain(**constraints).get_download_script()
        :return: A string containing the script

        """
        if constraints:
            sc = self.constrain(**constraints)
        else:
            sc = self

        sc.__update_counts()

        query_dict = sc._build_query()

        #!TODO: allow setting limit
        script = sc.connection.send_wget(query_dict,
                                         shards=self.shards)

        return script

    @property
    def facet_counts(self):
        self.__update_counts()
        return self.__facet_counts

    @property
    def hit_count(self):
        self.__update_counts()
        return self.__hit_count

    def get_facet_options(self):
        """
        Return a dictionary of facet counts filtered to remove all
        facets that are completely constrained.  This method is
        similar to the property ``facet_counts`` except facet values
        which are not relevant for further constraining are removed.
        
        """
        facet_options = {}
        hits = self.hit_count
        for facet, counts in self.facet_counts.items():
            # filter out counts that match total hits
            counts = dict(items for items in counts.items()
                          if items[1] < hits)
            if len(counts) > 1:
                facet_options[facet] = counts

        return facet_options

    def __update_counts(self):
        # If hit_count is set the counts are already retrieved
        if self.__hit_count is not None:
            return

        self.__facet_counts = {}
        self.__hit_count = None
        query_dict = self._build_query()
        query_dict['facets'] = '*'

        response = self.connection.send_search(query_dict, limit=0)
        for facet, counts in (
            response['facet_counts']['facet_fields'].items()):
            d = self.__facet_counts[facet] = {}
            while counts:
                d[counts.pop()] = counts.pop()

        self.__hit_count = response['response']['numFound']

    #-------------------------------------------------------------------------
    # Constraint mutation interface
    # These functions update the instance in-place.
    # Use constrain() and search() to generate new contexts with tighter
    # constraints.

    def _update_constraints(self, constraints):
        """
        Update the constraints in-place by calling _constrain_*() methods.
        
        """
        constraints_split = self._split_constraints(constraints)
        self._constrain_facets(constraints_split['facet'])
        if 'query' in constraints_split['freetext']:
            new_freetext = constraints_split['freetext']['query']
            self._constrain_freetext(new_freetext)

        #!TODO: implement temporal and geospatial constraints
        if 'from_timestamp' in constraints_split['temporal']:
            self.temporal_constraint[0] = constraints_split['temporal']['from_timestamp']
        if 'to_timestamp' in constraints_split['temporal']:
            self.temporal_constraint[1] = constraints_split['temporal']['to_timestamp']
        #self._constrain_geospatial()

        # reset cached values
        self.__hit_count = None
        self.__facet_counts = None

    def _constrain_facets(self, facet_constraints):
        for key, values in facet_constraints.mixed().items():
            current_values = self.facet_constraints.getall(key)
            if isinstance(values, list):
                for value in values:
                    if value not in current_values:
                        self.facet_constraints.add(key, value)
            else:
                if values not in current_values:
                    self.facet_constraints.add(key, values)

    def _constrain_freetext(self, query):
        self.freetext_constraint = query

    def _constrain_geospatial(self, lat=None, lon=None, bbox=None, location=None,
                              radius=None, polygon=None):
        self.geospatial_constraint = GeospatialConstraint(lat, lon, bbox, location, radius, polygon)

        raise NotImplementedError

    #-------------------------------------------------------------------------

    def _split_constraints(self, constraints):
        """
        Divide a constraint dictionary into 4 types of constraints:
        1. Freetext query
        2. Facet constraints
        3. Temporal constraints
        4. Geospatial constraints

        :return: A dictionary of the 4 types of constraint.
        
        """
        # local import to prevent circular importing
        from .connection import query_keyword_type

        constraints_split = dict((kw, MultiDict()) for kw in QUERY_KEYWORD_TYPES)

        for kw, val in constraints.items():
            constraint_type = query_keyword_type(kw)
            constraints_split[constraint_type][kw] = val

        return constraints_split

    def _build_query(self):
        """
        Build query string parameters as a dictionary.

        """

        query_dict = MultiDict({"query": self.freetext_constraint,
                                "type": self.search_type,
                                "latest": self.latest,
                                "facets": self.facets,
                                "fields": self.fields,
                                "replica": self.replica,
        })

        query_dict.extend(self.facet_constraints)

        #!TODO: encode datetime
        start, end = self.temporal_constraint
        query_dict.update(start=start, end=end)

        return query_dict


class DatasetSearchContext(SearchContext):
    DEFAULT_SEARCH_TYPE = TYPE_DATASET


class FileSearchContext(SearchContext):
    DEFAULT_SEARCH_TYPE = TYPE_FILE


class AggregationSearchContext(SearchContext):
    DEFAULT_SEARCH_TYPE = TYPE_AGGREGATION




"""

Module :mod:`pyesgf.logon`
==========================

Manage the client's interaction with ESGF's security system.  Using this module requires installing the MyProxyClient_ library.

.. _MyProxyClient: http://pypi.python.org/pypi/MyProxyClient

To obtain ESGF credentials create a :class:`LogonManager` instance and supply it with logon
details::

  >>> lm = LogonManager()
  >>> lm.is_logged_on()
  False
  >>> lm.logon(username, password, myproxy_hostname)
  >>> lm.is_logged_on()
  True

Logon parameters that aren't specified will be prompted for at the terminal
by default.  The :class:`LogonManager` object also writes a ``.httprc`` file configuring
OPeNDAP access through the NetCDF API.

You can use your OpenID to logon instead.  The logon details will be deduced
from the OpenID where possible::

  >>> lm.logoff()
  >>> lm.is_logged_on()
  False
  >>> lm.logon_with_openid(openid, password)
  >>> lm.is_logged_on()
  True

"""

import os
import os.path as op
import datetime
import shutil
from xml.etree import ElementTree
import urllib2
import re
from getpass import getpass

try:
    from myproxy.client import MyProxyClient
    import OpenSSL
    _has_myproxy = True
except ImportError:
    _has_myproxy = False

from .exceptions import OpenidResolutionError

#-----------------------------------------------------------------------------
# Constants

ESGF_DIR = op.join(os.environ['HOME'], '.esg')
ESGF_CERTS_DIR = 'certificates'
ESGF_CREDENTIALS = 'credentials.pem'
DAP_CONFIG = op.join(os.environ['HOME'], '.dodsrc')
DAP_CONFIG_MARKER = '<<< Managed by esgf-pyclient >>>'

XRI_NS = 'xri://$xrd*($v*2.0)'
MYPROXY_URN = 'urn:esg:security:myproxy-service'
ESGF_OPENID_REXP = r'https://.*/esgf-idp/openid/(.*)'
MYPROXY_URI_REXP = r'socket://([^:]*):?(\d+)?'

#-----------------------------------------------------------------------------
# classes

class LogonManager(object):
    """
    Manages ESGF crendentials and security configuration files.

    Also integrates with NetCDF's secure OPeNDAP configuration.
    
    """
    STATE_LOGGED_ON = 0
    STATE_NO_CREDENTIALS = 1
    STATE_EXPIRED_CREDENTIALS = 2
    STATE_INVALID_CREDENTIALS = 3

    def __init__(self, esgf_dir=ESGF_DIR, dap_config=DAP_CONFIG):
        """
        :param esgf_dir: Root directory of ESGF state.  Default ~/.esg
        :param dap_config: Set the location of .httprc.  Defaults to ~/.httprc

        Note if dap_config is defined your current working directory must be the
        same as the location as the dap_config file when OPeNDAP is initiallised.

        """
        if not _has_myproxy:
            raise ImportError('pyesgf.logon requires MyProxyClient')

        self.esgf_dir = esgf_dir
        self.esgf_credentials = op.join(self.esgf_dir, ESGF_CREDENTIALS)
        self.esgf_certs_dir = op.join(self.esgf_dir, ESGF_CERTS_DIR)
        self.dap_config = dap_config

        self._write_dap_config()

    @property
    def state(self):
        if not op.exists(self.esgf_credentials):
            return self.STATE_NO_CREDENTIALS
        else:
            with open(self.esgf_credentials) as fh:
                data = fh.read()
                cert = OpenSSL.crypto.load_certificate(OpenSSL.SSL.FILETYPE_PEM, data)

            if cert.has_expired():
                return self.STATE_EXPIRED_CREDENTIALS

            #!TODO: check credentials against certificates

            return self.STATE_LOGGED_ON


    def is_logged_on(self):
        return self.state == self.STATE_LOGGED_ON


    def logon_with_openid(self, openid, password=None,
                          bootstrap=False, update_trustroots=True,
                          interactive=True):
        """
        Obtains ESGF credentials by detecting the MyProxy parameters from
        the users OpenID.  Some ESGF compatible OpenIDs do not contain enough
        information to obtain credentials.  In this case the user is prompted
        for missing information if ``interactive == True``, otherwise an
        exception is raised.

        :param openid: OpenID to login with See :meth:`logon` for parameters
            ``interactive``, ``bootstrap`` and ``update_trustroots``.
        """
        username, myproxy = self._get_logon_details(openid)
        return self.logon(username, password, myproxy,
                          bootstrap=bootstrap,
                          update_trustroots=update_trustroots,
                          interactive=interactive)


    def logon(self, username=None, password=None, hostname=None,
              bootstrap=False, update_trustroots=True,
              interactive=True):
        """
        Obtain ESGF credentials from the specified MyProxy service.

        If ``interactive == True`` then any missing parameters of ``password``,
        ``username`` or ``hostname`` will be prompted for at the terminal.
        
        :param interactive: Whether to ask for input at the terminal for
            any missing information.  I.e. username, password or hostname.
        :param bootstrap: Whether to bootstrap the trustroots for this
            MyProxy service.
        :param update_trustroots: Whether to update the trustroots for this
            MyProxy service.

        """
        if interactive:
            if hostname is None:
                print 'Enter myproxy hostname:',
                hostname = raw_input()
            if username is None:
                print 'Enter myproxy username:',
                username = raw_input()
            if password is None:
                password = getpass('Enter password for %s: ' % username)

        if None in (hostname, username, password):
            raise OpenidResolutionError('Full logon details not available')

        c = MyProxyClient(hostname=hostname, caCertDir=self.esgf_certs_dir)

        creds = c.logon(username, password,
                        bootstrap=bootstrap, updateTrustRoots=update_trustroots)
        with open(self.esgf_credentials, 'w') as fh:
            for cred in creds:
                fh.write(cred)
                

    def logoff(self, clear_trustroots=False):
        """
        Remove any obtained credentials from the ESGF environment.
        
        :param clear_trustroots: If True also remove trustroots.
    
        """
        if op.exists(self.esgf_credentials):
            os.remove(self.esgf_credentials)

        if clear_trustroots:
            shutil.rmtree(self.esgf_certs_dir)

    def _get_logon_details(self, openid):
        openid_doc = urllib2.urlopen(openid).read()
        xml = ElementTree.fromstring(openid_doc)

        hostname = None
        port = None
        username = None

        services = xml.findall('.//{%s}Service' % XRI_NS)
        for service in services:
            try:
                service_type = service.find('{%s}Type' % XRI_NS).text
            except AttributeError:
                continue

            # Detect myproxy hostname and port
            if service_type == MYPROXY_URN:
                myproxy_uri = service.find('{%s}URI' % XRI_NS).text
                mo = re.match(MYPROXY_URI_REXP, myproxy_uri)
                if mo:
                    hostname, port = mo.groups()

        # If the OpenID matches the standard ESGF pattern assume it contains
        # the username, otherwise prompt or raise an exception
        mo = re.match(ESGF_OPENID_REXP, openid)
        if mo:
            username = mo.group(1)
                
        #!TODO maybe support different myproxy port
        if port is not None:
            assert int(port) == 7512
            
        return username, hostname
            

    def _write_dap_config(self, verbose=False, validate=False):
        preamble, managed, postamble = self._parse_dap_config()

        with open(self.dap_config, 'w') as fh:
            fh.write("""\
{preamble}
# BEGIN {marker}
HTTP.VERBOSE={verbose}
HTTP.COOKIEJAR={esgf_dir}/.dods_cookies
HTTP.SSL.VALIDATE=0
HTTP.SSL.CERTIFICATE={esgf_dir}/credentials.pem
HTTP.SSL.KEY={esgf_dir}/credentials.pem
HTTP.SSL.CAPATH={esgf_certs_dir}
# END {marker}
{postamble}
""".format(verbose=1 if verbose else 0, 
           validate=1 if validate else 0,
           esgf_certs_dir=self.esgf_certs_dir, 
           esgf_dir=self.esgf_dir,
           marker=DAP_CONFIG_MARKER,
           preamble=preamble,
           postamble=postamble,
           ))


    def _parse_dap_config(self, config_str=None):
        """
        Read the DAP_CONFIG file and extract the parts not controlled 
        by esgf-pyclient.

        :return: (preamble, managed, postamble), three strings of 
            configuration lines before, within and after the esgf-pyclient 
            controlled block.

        """
        if config_str is None:
            if not op.exists(self.dap_config):
                return ('', '', '')
            config_str = open(self.dap_config).read()

        #!NOTE: The flags keyword argument to re.split was introduced in Python2.7
        #   Keep with call non-keyword arguments for compatibility with Python2.6
        sections = re.split(r'^# (?:BEGIN|END) {0}$\n'.format(DAP_CONFIG_MARKER), 
                            config_str, re.M)

        if len(sections) < 2:
            preamble, managed, postamble = sections[0], '', ''
        elif len(sections) == 2:
            preamble, managed, postamble = sections + ['']
        elif len(sections) == 3:
            preamble, managed, postamble = sections
        else:
            # In odd circumstances there might be more than 3 parts of the config 
            # so assume the final config is the one to keep
            managed, unmanaged = [], []
            sections.reverse()
            while sections:
                unmanaged.append(sections.pop())
                if sections:
                    managed.append(sections.pop())

            preamble = '\n'.join(unmanaged[:-1])
            postamble = unmanaged[-1]
            managed = managed[-1]


        return preamble.strip(), managed.strip(), postamble.strip()



"""
Generate ESGF manifest files from ESGF search results

!TODO: Manifest specification

"""

import os
import urllib2
import csv
import re
import datetime

from pyesgf.exceptions import Error, DuplicateHashError

import logging
log = logging.getLogger(__name__)

class Manifest(object):
    VERSION = '0.1'

    def __init__(self, drs_id):
        self.drs_id = drs_id
        self._contents = {}

    @classmethod
    def from_result(cls, dataset_result):
        """
        Instantiate a Manifest instance from a :class:`search.results:DatasetResult`.
        
        """

        ctx = dataset_result.file_context()
        obj = cls()
        for file_result in ctx.search():
            filename = file_result.filename
            #!TODO: trap exceptions here
            filehash = '%s:%s' % (file_result.checksum_type.lower(), 
                                  file_result.checksum)
            tracking_id = file_result.tracking_id
            if tracking_id is None:
                tracking_id = ''
            size = file_result.size

            obj.add(filename, filehash, tracking_id, size)

        return obj

    @classmethod
    def from_mapfile(cls, mapfile):
        #!TODO
        raise NotImplementedError


    def add(self, filename, filehash, tracking_id, size):
        if filename in self._contents:
            old_filehash, old_tracking_id, old_size = self._contents[filename]
            if old_filehash == filehash:
                log.warn('Filename {0} occurs multiple times in dataset {1} '
                         'with the same hash'.format(filename, self.drs_id))
            elif old_tracking_id == tracking_id:
                raise DuplicateHashError(
                    'Filename {0} occurs multiple times in dataset {1} with '
                    'the same tracking_id and different hash'.format(filename,
                                                                     self.drs_id))
            else:
                raise DuplicateHashError(
                    'Filename {0} occurs multiple times in '
                    'dataset {1}'.format(filename, self.drs_id))
            
        self._contents[filename] = (filehash, tracking_id, size)


    def write(self, fh):
        fh.write('#%esgf-manifest {0}\n'.format(self.VERSION))
        for filename in sorted(self._contents):
            filehash, tracking_id, size = self._contents[filename]
            fh.write('{0},{1},{2},{3}'.format(filename, filehash, tracking_id, size)+'\n')



class ManifestExtractor(object):
    pass


class SolrManifestExtractor(ManifestExtractor):
    SOLR_BATCH_SIZE = 1000
    SOLR_FIELDS = ['dataset_id', 'title', 'checksum_type', 'checksum',
                   'tracking_id','size']

    def __init__(self, endpoint, project, from_date=None):
        self.endpoint = endpoint
        self.project = project
        self.from_date = from_date

    def _query(self, offset):
        fq = 'project:{0}'.format(self.project)
        if self.from_date:
            fq += ' AND timestamp[{0}Z TO NOW]'.format(self.from_date)

        params = (
            ('q', '*'),
            ('fq', fq),
            ('fl', ','.join(self.SOLR_FIELDS)),
            ('wt', 'csv'),
            ('sort', 'dataset_id%20asc'),
            ('start', offset),
            ('rows', self.SOLR_BATCH_SIZE),
            )
        param_str = '&'.join(('{0}={1}'.format(k, v)) for (k, v) in params)

        url = '{0}/solr/files/select?{1}'.format(self.endpoint,
                                                 param_str)
        log.info('SOLR QUERY: {0}'.format(url))
        response = urllib2.urlopen(url)
        
        return response

    def _check_header(self, header):
        if header != self.SOLR_FIELDS:
            raise Error('Unrecognised SOLr CSV header')


    def _init_manifest(self, dataset_id):
        drs_id, datanode = dataset_id.split('|')
        log.info('Generating manifest {0}'.format(drs_id))

        return Manifest(drs_id)


    def __iter__(self):
        """
        Query the SOLr index until all files for this project have been extracted.
 
        """

        offset = 0
        current_dataset_id = None
        current_manifest = None
        while 1:
            response = self._query(offset)
            reader = csv.reader(response)

            # Read first line and check the header
            header = reader.next()
            self._check_header(header)

            empty_batch = True
            for row in reader:
                empty_batch = False

                log.debug('ROW: {0}'.format(row))
                dataset_id, filename, checksum_type, checksum, tracking_id, size = row
                if dataset_id != current_dataset_id:
                    if current_manifest:
                        yield current_manifest
                    current_manifest = self._init_manifest(dataset_id)                    
                    current_dataset_id = dataset_id

                filehash = '{0}:{1}'.format(checksum_type.lower(), checksum)
                size = int(size)

                try:
                    current_manifest.add(filename, filehash, tracking_id, size)
                except DuplicateHashError, e:
                    log.error(e)

            # Test whether there were any rows.  If not we have reached the end.
            if empty_batch:
                break
            else:
                offset += self.SOLR_BATCH_SIZE

        if current_manifest:
            yield current_manifest


def cmip5_manifest_partitioner(drs_id):
    """
    Split a drs_id into pragmatic components for constructing a directory structure.
    This partition is designed to balance the need to split a repository between
    subdirectories with the desire to minimise the depth of the tree.

    """
    (activity, product, institute, model, experiment, 
     frequency, realm, table, ensemble, version) = drs_id.split('.')

    parts = ('.'.join((activity, product, institute)),
             '.'.join((model, experiment, frequency)))
    
    return parts



def extract_from_solr(endpoint, project, target_dir, from_date=None):
    solr_extractor = SolrManifestExtractor(endpoint, project, from_date)

    for manifest in solr_extractor:

        manifest_path = os.path.join(target_dir,
                                     *cmip5_manifest_partitioner(manifest.drs_id))
        if not os.path.exists(manifest_path):
            log.info('Creating repo directory {0}'.format(manifest_path))
            os.makedirs(manifest_path)
        
        manifest_file = os.path.join(manifest_path, manifest.drs_id)


        if os.path.exists(manifest_file):
            log.warn('Manifest {0} already exists.  This manifest will be overwritten'.format(manifest_file))

        with open(manifest_file, 'w') as fh:
            log.info('Writing Manifest {0}'.format(manifest_file))
            manifest.write(fh)


def parse_timestamp(timestamp):
    """
    Parses a standard ISO timestamp.  This function allows the time to be ommitted but
    if present it must include hours, minutes and seconds.

    """
    if re.match(r'\d{4}-\d{2}-\d{2}$', timestamp):
        timestamp += 'T00:00:00Z'
    elif timestamp[-1] != 'Z':
        timestamp += 'Z'

    dt = datetime.datetime.strptime(timestamp, '%Y-%m-%dT%H:%M:%SZ')

    return dt

if __name__ == '__main__':
    import argparse

    logging.basicConfig(level=logging.INFO)

    parser = argparse.ArgumentParser(description='Extract ESGF datasets from a SOLr index into a manifest repository')
    parser.add_argument('endpoint', help='SOLr endpoint as http://<host>:<port>')
    parser.add_argument('project', help='The project to extract')
    parser.add_argument('repository', help='Path to the manifest repository')
    parser.add_argument('--from', dest='from_date', action='store', 
                        help='The minimum timestamp of returned SOLr records')

    args = parser.parse_args()

    if args.from_date:
        from_date = parse_timestamp(args.from_date)
    else:
        from_date is None

    extract_from_solr(args.endpoint, args.project, args.repository, from_date=from_date)



"""
Test pyesgf.util module.

"""

import re

from pyesgf.search.connection import SearchConnection
from pyesgf.util import get_manifest

#!TODO: Make this one of the test services.
CEDA_SERVICE = 'http://esgf-index1.ceda.ac.uk/esg-search'

def test_get_manifest():
    conn = SearchConnection(CEDA_SERVICE, distrib=False)

    manifest = get_manifest('GeoMIP.output1.MOHC.HadGEM2-ES.G1.day.atmos.day.r1i1p1',
                         20120223, conn)

    filename = 'psl_day_HadGEM2-ES_G1_r1i1p1_19291201-19291230.nc'
    assert manifest[filename]['checksum'] == 'd20bbba8e05d6689f44cf3f8eebb9e7b'

#!TODO: this test belongs somewhere else
def test_opendap_url():
    conn = SearchConnection(CEDA_SERVICE, distrib=False)

    ctx = conn.new_context()
    results = ctx.search(drs_id='GeoMIP.output1.MOHC.HadGEM2-ES.G1.day.atmos.day.r1i1p1')
    assert len(results) == 1

    agg_ctx = results[0].aggregation_context()
    aggs = agg_ctx.search()

    # Take first aggregation
    agg = aggs[0]
     
    print agg.aggregation_id
    print agg.json['cf_standard_name']
    print agg.urls

    opendap_url = agg.opendap_url
    print opendap_url

def test_download_url():
    conn = SearchConnection(CEDA_SERVICE, distrib=False)

    ctx = conn.new_context()
    results = ctx.search(drs_id='GeoMIP.output1.MOHC.HadGEM2-ES.G1.day.atmos.day.r1i1p1')
    files = results[0].file_context().search()

    download_url = files[0].download_url
    assert re.match(r'http://.*\.nc', download_url)
    

def test_opendap_fail():
    conn = SearchConnection(CEDA_SERVICE, distrib=False)

    ctx = conn.new_context()
    results = ctx.search(project='CMIP5', experiment='rcp45', time_frequency='mon',
                         realm='atmos', ensemble='r1i1p1')

    files_ctx = results[0].file_context()
    hit = files_ctx.search()[0]

    assert hit.opendap_url is None



"""

Module :mod:`pyesgf.search.results`
===================================

Search results are retrieved through the :class:`ResultSet` class.  This class
hides paging of large result sets behind a client-side cache.  Subclasses of 
:class:`Result` represent results of different SOLr record type.

"""

from collections import Sequence, defaultdict
import re

from .consts import (DEFAULT_BATCH_SIZE, TYPE_DATASET, TYPE_FILE, 
                     TYPE_AGGREGATION)
from .exceptions import EsgfSearchException

class ResultSet(Sequence):
    """
    :ivar context: The search context object used to generate this resultset
    :property batch_size: The number of results that will be requested
        from esgf-search as one call.  This must be set on creation and
        cannot change.

    """
    def __init__(self, context, batch_size=DEFAULT_BATCH_SIZE, eager=True):
        """
        :param context: The search context object used to generate this resultset
        :param batch_size: The number of results that will be requested from
            esgf-search as one call.
        :param eager: Boolean specifying whether to retrieve the first batch on
            instantiation.
            
        """
        self.context = context
        self.__batch_size = batch_size
        self.__batch_cache = [None] * ((len(self) / batch_size) + 1)
        if eager and len(self)>0:
            self.__batch_cache[0] = self.__get_batch(0)

    def __getitem__(self, index):
        batch_i = index / self.batch_size
        offset = index % self.batch_size
        if self.__batch_cache[batch_i] is None:
            self.__batch_cache[batch_i] = self.__get_batch(batch_i)
        batch = self.__batch_cache[batch_i]

        search_type = self.context.search_type
        ResultClass = _result_classes[search_type]

        #!TODO: should probably wrap the json inside self.__batch_cache
        return ResultClass(batch[offset], self.context)
            

    def __len__(self):
        return self.context.hit_count

    @property
    def batch_size(self):
        return self.__batch_size

    def _build_result(self, result):
        """
        Construct a result object from the raw json.

        This method is designed to be overridden in subclasses if desired.
        The default implementation simply returns the json.
        
        """
        return result

    def __get_batch(self, batch_i):
        offset = self.batch_size * batch_i
        limit = self.batch_size

        query_dict = self.context._build_query()
        response = self.context.connection.send_search(query_dict, limit=limit, 
                                                       offset=offset,
                                                       shards=self.context.shards)
        
        #!TODO: strip out results
        return response['response']['docs']


class BaseResult(object):
    """
    Base class for results.

    Subclasses represent different search types such as File and Dataset.

    :ivar json: The oroginial json representation of the result.
    :ivar context: The SearchContext which generated this result.
    :property urls: a dictionary of the form {service: [(url, mime_type), ...], ...}
    :property opendap_url: The url of an OPeNDAP endpoint for this result if available
    :property download_url: The url for downloading the result by HTTP if available
    :property index_node: The index node from where the metadata is stored.  
        Calls to *_context() will optimise queries to only address this node.

    """
    def __init__(self, json, context):
        self.json = json
        self.context = context
        
    @property
    def urls(self):
        url_dict = defaultdict(list)
        for encoded in self.json['url']:
            url, mime_type, service = encoded.split('|')
            url_dict[service].append((url, mime_type))

        return url_dict

    @property
    def opendap_url(self):
        try:
            url, mime = self.urls['OPENDAP'][0]
        except (KeyError, IndexError):
            return None
        
        url = re.sub(r'.html$', '', url)

        return url

    @property
    def las_url(self):
        try:
            url, mime = self.urls['LAS'][0]
        except (KeyError, IndexError):
            return None

        return url

    @property
    def download_url(self):
        try:
            url, mime = self.urls['HTTPServer'][0]
        except (KeyError, IndexError):
            return None

        return url

    @property
    def index_node(self):
        try:
            index_node = self.json['index_node']
        except KeyError:
            return None

        return index_node


class DatasetResult(BaseResult):
    """
    A result object for ESGF datasets.

    :property dataset_id: The solr dataset_id which is unique throughout the system.

    """

    @property
    def dataset_id(self):
        #!TODO: should we decode this into a tuple?  self.json['id'].split('|')
        return self.json['id']
    
    @property
    def number_of_files(self):
        """
        Returns file count as reported by the dataset record.
        """
        return self.json['number_of_files']

    def file_context(self):
        """
        Return a SearchContext for searching for files within this dataset.
        """
        from .context import FileSearchContext

        if self.context.connection.distrib:
            # If the index node is in the available shards for this connection then
            # restrict shards to that node.  Otherwise do nothing to handle the case
            # when the shard is replicated
            available_shards = self.context.connection.get_shard_list().keys()
            if self.index_node in available_shards:
                shards = [self.index_node]
            else:
                shards = None
        else:
            shards = None

        files_context = FileSearchContext(
            connection=self.context.connection,
            constraints={'dataset_id': self.dataset_id},
            shards=shards,
            )
        return files_context

    def aggregation_context(self):
        """
        Return a SearchContext for searching for aggregations within this dataset.
        """
        from .context import AggregationSearchContext

        if self.context.connection.distrib:
            # If the index node is in the available shards for this connection then
            # restrict shards to that node.  Otherwise do nothing to handle the case
            # when the shard is replicated
            available_shards = self.context.connection.get_shard_list().keys()
            if self.index_node in available_shards:
                shards = [self.index_node]
            else:
                shards = None
        else:
            shards = None

        agg_context = AggregationSearchContext(
            connection=self.context.connection,
            constraints={'dataset_id': self.dataset_id},
            shards=shards,
            )
        return agg_context

class FileResult(BaseResult):
    """
    A result object for ESGF files.  Properties from :class:`BaseResult` are inherited.

    :property file_id: The identifier for the file
    :property checksum: The checksum of the file
    :property checksum_type: The algorithm used for generating the checksum
    :property filename: The filename
    :proprty size: The file size in bytes

    """
    @property
    def file_id(self):
        return self.json['id']

    @property
    def checksum(self):
        try:
            return self.json['checksum'][0]
        except KeyError:
            return None

    @property
    def checksum_type(self):
        try:
            return self.json['checksum_type'][0]
        except KeyError:
            return None

    @property
    def filename(self):
        return self.json['title']

    @property
    def size(self):
        return int(self.json['size'])

    @property
    def tracking_id(self):
        try:
            return self.json['tracking_id'][0]
        except KeyError:
            return None


class AggregationResult(BaseResult):
    """
    A result object for ESGF aggregations.  Properties from :class:`BaseResult` are inherited.

    :property aggregation_id: The aggregation id
    """
    @property
    def aggregation_id(self):
        return self.json['id']


_result_classes = {
    TYPE_DATASET: DatasetResult,
    TYPE_FILE: FileResult,
    TYPE_AGGREGATION: AggregationResult,
    }



"""
Top-level package for Python interface to ESGF services.

"""

__version__ = '0.1.4b'

#!TODO: ResultFormatter class.  process response json to specialise the result json.  Default is None
#!TODO: pipe results to new process.  Command-line interface.
#!TODO: Helper methods for "get opendap" "get download urls"

    
    

    



"""
Utility functions using the pyesgf package.

"""

import sys
from urllib import quote_plus, _is_unicode

def get_manifest(drs_id, version, connection):
    """
    Retrieve the filenames, sizes and checksums of a dataset.

    This function will raise ValueError if more than one dataset is found
    matching the given drs_id and version on a search without replicas.
    The connection should be either distrib=True or be connected to a suitable
    ESGF search interface.

    :param drs_id: a string containing the DRS identifier without version
    :param version: The version as a string or int

    """

    if type(version) == int:
        version = str(version)

    context = connection.new_context(drs_id=drs_id, version=version)
    results = context.search()

    if len(results) > 1:
        raise ValueError("Search for dataset %s.v%s returns multiple hits" % 
                         (drs_id, version))

    file_context = results[0].file_context()
    manifest = {}
    for file in file_context.search():
        manifest[file.filename] = {
                'checksum_type': file.checksum_type,
                'checksum': file.checksum,
                'size': file.size,
                }

    return manifest
            

def urlencode(query):
    """Encode a sequence of two-element tuples or dictionary into a URL query string.

    This version is adapted from the standard library to understand operators in the
    pyesgf.search.constraints module.

    If the query arg is a sequence of two-element tuples, the order of the
    parameters in the output will match the order of parameters in the
    input.
    """

    if hasattr(query,"items"):
        # mapping objects
        query = query.items()
    else:
        # it's a bother at times that strings and string-like objects are
        # sequences...
        try:
            # non-sequence items should not work with len()
            # non-empty strings will fail this
            if len(query) and not isinstance(query[0], tuple):
                raise TypeError
            # zero-length sequences of all types will get here and succeed,
            # but that's a minor nit - since the original implementation
            # allowed empty dicts that type of behavior probably should be
            # preserved for consistency
        except TypeError:
            ty,va,tb = sys.exc_info()
            raise TypeError, "not a valid non-string sequence or mapping object", tb


    def append(k, v, tag, l):
        from .search.consts import OPERATOR_NEQ

        if tag == OPERATOR_NEQ:
            l.append('%s!=%s' % (k, v))
        elif tag is None:
            l.append('%s=%s' % (k, v))
        else:
            raise ValueError('Unknown operator tag %s' % tag)

    def strip_tag(v):
        if type(v) == tuple:
            tag, v = v
        else:
            tag = None
            
        return tag, v

    l = []
    for k, v in query:
        tag, v = strip_tag(v)
        k = quote_plus(str(k))
        if isinstance(v, str):
            v = quote_plus(v)
            append(k, v, tag, l)
        elif _is_unicode(v):
            # is there a reasonable way to convert to ASCII?
            # encode generates a string, but "replace" or "ignore"
            # lose information and "strict" can raise UnicodeError
            v = quote_plus(v.encode("ASCII","replace"))
            append(k, v, tag, l)
        else:
            try:
                # is this a sufficient test for sequence-ness?
                len(v)
            except TypeError:
                # not a sequence
                v = quote_plus(str(v))
                append(k, v, tag, l)
            else:
                # loop over the sequence
                for elt in v:
                    append(k, quote_plus(str(elt)), tag, l)
    return '&'.join(l)



# BSD Licence
# Copyright (c) 2012, Science & Technology Facilities Council (STFC)
# All rights reserved.
#
# See the LICENSE file in the source distribution of this software for
# the full license text.

from setuptools import setup, find_packages
import sys, os


# Import version from the top-level package
sys.path[:0] = os.path.dirname(__file__)
from pyesgf import __version__
from pyesgf import __doc__ as long_description

setup(name='esgf-pyclient',
      version=__version__,
      description="A library interacting with ESGF services within Python",
      long_description=long_description,
      classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: BSD License',
        'Topic :: Scientific/Engineering',
        'Programming Language :: Python :: 2.6',
        ], # Get strings from http://pypi.python.org/pypi?%3Aaction=list_classifiers
      keywords='',
      author='Ag Stephens',
      author_email='Ag.Stephens@stfc.ac.uk',
      url='http://esgf-pyclient.readthedocs.org',
      download_url='http://github.com/ESGF/esgf-pyclient',
      license='BSD',
      packages=find_packages(exclude=['ez_setup', 'examples', 'test']),
      include_package_data=True,
      zip_safe=False,
      install_requires=[
        'jinja2',
      ],
      #tests_require=['NoseXUnit'],
      entry_points= {
        },
      test_suite='nose.collector',
      )



"""

Module :mod:`pyesgf.search.connection`
======================================

Defines the class representing connections to the ESGF Search API.  To
perform a search create a :class:`SearchConnection` instance then use
:meth:`new_context()` to create a search context.

.. warning:: 
   Prior to v0.1.1 the *url* parameter expected the full URL of the
   search endpoint up to the query string.  This has now been changed
   to expect *url* to ommit the final endpoint name,
   e.g. ``http://pcmdi9.llnl.gov/esg-search/search`` should be changed
   to ``http://pcmdi9.llnl.gov/esg-search`` in client code.  The
   current implementation detects the presence of ``/search`` and
   corrects the URL to retain backward compatibility but this feature
   may not remain in future versions.
        
"""

import urllib2
import json
import re
from urlparse import urlparse

import warnings
import logging

logging.basicConfig()
log = logging.getLogger(__name__)
log.setLevel(logging.INFO)

from .context import DatasetSearchContext
from .consts import RESPONSE_FORMAT, SHARD_REXP
from .exceptions import EsgfSearchException
from pyesgf.multidict import MultiDict
from pyesgf.util import urlencode


class SearchConnection(object):
    """
    :ivar url: The URL to the Search API service.  This should be the URL 
        of the ESGF search service excluding the final endpoint name.  
        Usually this is http://<hostname>/esg-search
    :ivar distrib: Boolean stating whether searches through this connection are
        distributed.  I.e. whether the Search service distributes the query to
        other search peers.

    """
    # Default limit for queries.  None means use service default.
    default_limit = None

    def __init__(self, url, distrib=True, context_class=None):
        """
        :param context_class: Override the default SearchContext class.

        """
        self.url = url
        self.distrib = distrib

        # Check URL for backward compatibility
        self.__check_url()

        # _available_shards stores all available shards once retrieved from the server.
        # A value of None means they haven't been retrieved yet.
        # Once set it is a dictionary {'host': [(port, suffix), ...], ...}
        self._available_shards = None

        if context_class:
            self.__context_class = context_class
        else:
            self.__context_class = DatasetSearchContext

    def __check_url(self):
        """
        Previous versions of the API expected the full URL to be given
        to SearchConnection's constructure.  This has been deprecated
        in favour of the URL without the final endpoint name in order
        to support other endpoints such as wget.

        This method tests whether self.url looks like an old-style full url and
        fixes the attribute accordingly, raising a depracation warning.
        
        """

        mo = re.match(r'(.*?)(/search)?/*$', self.url)
        assert mo

        if mo.group(2):
            warnings.warn('Old-style SearchContext URL specified.  '
                          'In future please specify the URL excluding '
                          'the "/search" endpoint.')

        self.url = mo.group(1)

    def send_search(self, query_dict, limit=None, offset=None, shards=None):
        """
        Send a query to the "search" endpoint.  See :meth:`send_query()` for details.

        :return: The json document for the search results

        """
        full_query = self._build_query(query_dict, limit, offset, shards)
        response = self._send_query('search', full_query)
        ret = json.load(response)

        return ret

    def send_wget(self, query_dict, shards=None):
        """
        Send a query to the "search" endpoint.  See :meth:`send_query()` for details.

        :return: A string containing the script.

        """
        full_query = self._build_query(query_dict, shards=shards)
        if 'type' in full_query:
            del full_query['type']
        if 'format' in full_query:
            del full_query['format']

        response = self._send_query('wget', full_query)
        script = response.read()

        return script

    def _send_query(self, endpoint, full_query):
        """
        Generally not to be called directly by the user but via SearchContext
    instances.
        
        :param full_query: dictionary of query string parameers to send.
        :return: the urllib2 response object from the query.
        
        """

        log.debug('Query dict is %s' % full_query)

        query_url = '%s/%s?%s' % (self.url, endpoint, urlencode(full_query))
        log.debug('Query request is %s' % query_url)

        try:
            response = urllib2.urlopen(query_url)
        except urllib2.HTTPError, err:
            log.warn("HTTP request received error code: %s" % err.code)
            if err.code == 400:
                errors = set(re.findall("Invalid HTTP query parameter=(\w+)", err.fp.read()))
                content = "; ".join([e for e in list(errors)])
                raise Exception("Invalid query parameter(s): %s" % content)
            else:
                raise Exception("Error returned from URL: %s" % query_url)

        return response


    def _build_query(self, query_dict, limit=None, offset=None, shards=None):
        if shards is not None:
            if self._available_shards is None:
                self._load_available_shards()

            shard_specs = []
            for shard in shards:
                if shard not in self._available_shards:
                    raise EsgfSearchException('Shard %s is not available' % shard)
                else:
                    for port, suffix in self._available_shards[shard]:
                        # suffix should be ommited when querying
                        shard_specs.append('%s:%s/solr' % (shard, port))

            shard_str = ','.join(shard_specs)
        else:
            shard_str = None

        full_query = MultiDict({
            'format': RESPONSE_FORMAT,
            'limit': limit,
            'distrib': 'true' if self.distrib else 'false',
            'offset': offset,
            'shards': shard_str,
        })
        full_query.extend(query_dict)

        # Remove all None valued items
        full_query = MultiDict(item for item in full_query.items() if item[1] is not None)

        return full_query

    def _load_available_shards(self):

        # Shards are not available if distrib=False.  The server won't send
        # back a list of shards
        if not self.distrib:
            raise EsgfSearchException('Shard list not available for '
                                      'non-distributed queries')

        self._available_shards = {}

        response_json = self.send_search({'facets': [], 'fields': []})

        try:
            shards = response_json['responseHeader']['params']['shards'].split(',')
        except KeyError:
            log.debug('_load_available_shards() fails with this exception')
            log.debug('response_json = %s' % response_json)
            raise EsgfSearchException('Error loading available shards')

        # Extract hostname and port from each shard.
        for shard in shards:
            mo = re.match(SHARD_REXP, shard)
            if not mo:
                raise EsgfSearchException('Shard spec %s not recognised' %
                                          shard)
            shard_parts = mo.groupdict()

            # Fix the host if it refers to the local server
            if shard_parts['host'] in ['localhost', '0.0.0.0', '127.0.0.1']:
                parsed_url = urlparse(self.url)
                shard_parts['host'] = parsed_url.hostname

            self._available_shards.setdefault(shard_parts['host'], []).append((shard_parts['port'],
                                                                               shard_parts['suffix']))


    def get_shard_list(self):
        """
        return the list of all available shards.  A subset of the returned list can be
        supplied to 'send_query()' to limit the query to selected shards.

        Shards are described by hostname and mapped to SOLr shard descriptions internally.

        :return: the list of available shards

        """
        if self._available_shards is None:
            self._load_available_shards()

        return self._available_shards


    def new_context(self, context_class=None,
                    latest=None, facets=None, fields=None,
                    from_timestamp=None, to_timestamp=None,
                    replica=None, shards=None,
                    **constraints):
        """
        Returns a :class:`pyesgf.search.context.SearchContext` class for 
        performing faceted searches.

        See :meth:`SearchContext.__init__()` for documentation on the 
        arguments.

        """
        if context_class is None:
            context_class = self.__context_class

        return context_class(self, constraints,
                             latest=latest, facets=facets, fields=fields,
                             from_timestamp=from_timestamp, 
                             to_timestamp=to_timestamp,
                             replica=replica, shards=shards,
        )


def query_keyword_type(keyword):
    """
    Returns the keyword type of a search query keyword.

    Possible values are 'system', 'freetext', 'facet', 'temporal' and
    'geospatial'.  If the keyword is unknown it is assumed to be a
    facet keyword

    """
    #!TODO: support "last update" constraints (to/from)

    if keyword == 'query':
        return 'freetext'
    elif keyword in ['start', 'end', 'from_timestamp', 'to_timestamp']:
        return 'temporal'
    elif keyword in ['lat', 'lon', 'bbox', 'location', 'radius', 'polygon']:
        return 'geospatial'
    elif keyword in ['limit', 'from', 'to', 'fields', 'facets', 'format',
                     'type', 'distrib', 'replica', 'id', 'shards']:
        return 'system'
    else:
        return 'facet'



"""
Interface to the ESGF SAML Attribute Service.

"""

from jinja2 import Template
import uuid
import datetime
import urllib2
from xml.etree import ElementTree as ET

from pyesgf.security import NS

import logging
log = logging.getLogger(__name__)

ATS_REQUEST_TMPL = Template('''<?xml version="1.0" encoding="UTF-8"?>
<soap11:Envelope xmlns:soap11="http://schemas.xmlsoap.org/soap/envelope/">
  <soap11:Body>
     <samlp:AttributeQuery xmlns:samlp="urn:oasis:names:tc:SAML:2.0:protocol" 
                           ID="{{ msg_id }}" 
                           IssueInstant="{{timestamp}}" Version="2.0">
        <saml:Issuer xmlns:saml="urn:oasis:names:tc:SAML:2.0:assertion" Format="urn:oasis:names:tc:SAML:1.1:nameid-format:X509SubjectName">{{ issuer }}</saml:Issuer>
        <saml:Subject xmlns:saml="urn:oasis:names:tc:SAML:2.0:assertion">
           <saml:NameID Format="urn:esg:openid">{{ openid }}</saml:NameID>
        </saml:Subject>
        {% for attr in attributes -%}
        <saml:Attribute xmlns:saml="urn:oasis:names:tc:SAML:2.0:assertion"
                        Name="{{ attr }}"
                        NameFormat="http://www.w3.org/2001/XMLSchema#string"/>
        {% endfor -%}
     </samlp:AttributeQuery>
  </soap11:Body>
</soap11:Envelope>
''')



class AttributeService(object):
    def __init__(self, url,issuer):
        self.url = url
        self.ISSUER=issuer

    def build_request(self, openid, attributes):
        now = datetime.datetime.utcnow()

        return ATS_REQUEST_TMPL.render(
            msg_id=uuid.uuid1(),
            timestamp=now.isoformat()+'Z',
            issuer=self.ISSUER,
            openid=openid,
            attributes=attributes)

    def send_request(self, openid, attributes):
        post_body = self.build_request(openid, attributes)
        req = urllib2.Request(self.url, post_body)
        req.add_header('Content-Type', 'text/xml')
        req.add_header('Content-Length', str(len(req.get_data())))
        log.debug(req.headers)
        log.debug(req.get_data())
        resp = urllib2.urlopen(req)

        return AttributeServiceResponse(resp)


class AttributeServiceResponse(object):
    def __init__(self, source):
        self.xml = ET.parse(source)

    def get_subject(self):
        fpath = './/{{{0}}}Subject/{{{0}}}NameID'.format(NS['saml'])
        subject_el = self.xml.find(fpath)

        return subject_el.text

    def get_attributes(self):
        d = {}
        fpath = './/{{{0}}}AttributeStatement/{{{0}}}Attribute'.format(NS['saml'])

        for el in self.xml.findall(fpath):
            attr_name = el.get('Name')
            val_el = el.find('./{{{0}}}AttributeValue'.format(NS['saml']))
            attr_value = val_el.text

            d.setdefault(attr_name, []).append(attr_value)

        # Remove list wrapper from single-valued attributes
        for key in d:
            if len(d[key]) == 1:
                d[key] = d[key][0]

        return d

    def get_status(self):
        fpath = './/{{{0}}}Status/{{{0}}}StatusCode'.format(NS['saml2p'])
        sc = self.xml.find(fpath)
        
        return sc.get('Value')
